import einops
import torch
import torch.nn.functional as F
from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)


class MMDiTDotProductAttention(nn.Module):
    def __init__(
            self,
            dim,
            num_modalities=2,
            num_heads=8,
            qkv_bias=True,
            proj_bias=True,
            channel_first=False,
            init_weights="truncnormal002",
            init_last_proj_zero=False,
    ):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.num_modalities = num_modalities
        self.channel_first = channel_first
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero

        self.qkv = nn.ModuleList(
            [
                nn.Linear(dim, dim * 3, bias=qkv_bias)
                for _ in range(num_modalities)
            ]
        )
        self.proj = nn.ModuleList(
            [
                nn.Linear(dim, dim, bias=proj_bias)
                for _ in range(num_modalities)
            ]
        )
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            for qkv in self.qkv:
                init_xavier_uniform_merged_linear(qkv, num_layers=3)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            for proj in self.proj:
                nn.init.zeros_(proj.weight)
                # init_weights == "torch" has no zero bias init
                if proj.bias is not None:
                    nn.init.zeros_(proj.bias)

    @staticmethod
    def to_channel_last(x):
        return einops.rearrange(x, "b c l -> b l c")

    @staticmethod
    def to_channel_first(x):
        return einops.rearrange(x, "b l c -> b c l")

    def forward(self, *args, attn_mask=None):
        x = list(args)
        assert len(x) == self.num_modalities
        assert all(x[i].ndim == 3 for i in range(self.num_modalities))
        if self.channel_first:
            for i in range(self.num_modalities):
                x[i] = self.to_channel_last(x[i])
        seqlens = [x[i].size(1) for i in range(self.num_modalities)]

        qs = []
        ks = []
        vs = []
        for i in range(self.num_modalities):
            q, k, v = einops.rearrange(
                self.qkv[i](x[i]),
                "bs seqlen (three num_heads head_dim) -> three bs num_heads seqlen head_dim",
                three=3,
                num_heads=self.num_heads,
                head_dim=self.head_dim,
            ).unbind(0)
            qs.append(q)
            ks.append(k)
            vs.append(v)
        # concat in sequence dimension
        q = torch.concat(qs, dim=2)
        k = torch.concat(ks, dim=2)
        v = torch.concat(vs, dim=2)
        # attention
        x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask)
        x = einops.rearrange(x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        x = list(x.split(seqlens, dim=1))
        for i in range(self.num_modalities):
            x[i] = self.proj[i](x[i])

        if self.channel_first:
            for i in range(self.num_modalities):
                x[i] = self.to_channel_first(x[i])
        return x



import einops
import torch
from torch import nn


class VitClassTokens(nn.Module):
    def __init__(
            self,
            dim: int,
            num_tokens: int = 1,
            location: str = "first",
            init_std: float = 0.02,
            init_std_pos: float = 0.02,
            aggregate: str = "flatten",
            decouple_position: bool = False,
    ):
        super().__init__()
        self.dim = dim
        self.location = location
        self.num_tokens = num_tokens
        self.init_std = init_std
        self.init_std_pos = init_std_pos
        self.aggregate = aggregate
        self.decouple_position = decouple_position
        if num_tokens > 0:
            if location in ["first", "middle", "last", "uniform"]:
                pass
            elif location == "bilateral":
                assert num_tokens % 2 == 0
            else:
                raise NotImplementedError
            self.tokens = nn.Parameter(torch.zeros(1, num_tokens, dim))
            if decouple_position:
                self.pos = nn.Parameter(torch.zeros(1, num_tokens, dim))
            else:
                self.pos = None
        else:
            self.tokens = None
            self.pos = None
        self.reset_parameters()

    def reset_parameters(self):
        if self.num_tokens > 0:
            nn.init.normal_(self.tokens, std=self.init_std)
            if self.decouple_position:
                nn.init.normal_(self.pos, std=self.init_std_pos)

    @property
    def tokens_proxy(self):
        if self.decouple_position:
            return self.tokens + self.pos
        return self.tokens

    def forward(self, x):
        if self.num_tokens == 0:
            return x
        assert x.ndim == 3
        if self.location == "first":
            tokens = self.tokens_proxy.expand(len(x), -1, -1)
            x = torch.concat([tokens, x], dim=1)
        elif self.location == "middle":
            tokens = self.tokens_proxy.expand(len(x), -1, -1)
            pre, post = x.chunk(chunks=2, dim=1)
            x = torch.concat([pre, tokens, post], dim=1)
        elif self.location == "last":
            tokens = self.tokens_proxy.expand(len(x), -1, -1)
            x = torch.concat([x, tokens], dim=1)
        elif self.location == "bilateral":
            first, last = self.tokens_proxy.chunk(chunks=2, dim=1)
            first = first.expand(len(x), -1, -1)
            last = last.expand(len(x), -1, -1)
            x = torch.concat([first, x, last], dim=1)
        elif self.location == "uniform":
            chunks = x.chunk(chunks=self.num_tokens + 1, dim=1)
            tokens = self.tokens_proxy.expand(len(x), -1, -1)
            # interweave chunk with token
            interweaved = [chunks[0]]
            for i in range(self.num_tokens):
                interweaved.append(tokens[:, i:i + 1])
                interweaved.append(chunks[i + 1])
            x = torch.concat(interweaved, dim=1)
        else:
            raise NotImplementedError
        return x

    @property
    def output_shape(self):
        if self.aggregate == "flatten":
            return self.dim * self.num_tokens,
        if self.aggregate == "mean":
            return self.dim,
        raise NotImplementedError

    def split(self, x):
        if self.num_tokens == 0:
            return None, x
        assert x.ndim == 3
        if self.location == "first":
            cls_tokens = x[:, :self.num_tokens]
            patch_tokens = x[:, self.num_tokens:]
        elif self.location == "middle":
            middle_start = (x.size(1) - self.num_tokens) // 2
            middle_end = middle_start + self.num_tokens
            cls_tokens = x[:, middle_start:middle_end]
            patch_tokens = torch.concat([x[:, :middle_start], x[:, middle_end:]], dim=1)
        elif self.location == "bilateral":
            cls_tokens = x[:, [0, -1]]
            patch_tokens = x[:, 1:-1]
        else:
            raise NotImplementedError
        return cls_tokens, patch_tokens

    def pool(self, x):
        if self.num_tokens == 0:
            raise NotImplementedError
        else:
            # extract tokens
            if self.location == "first":
                x = x[:, :self.num_tokens]
            elif self.location == "middle":
                middle = x.size(1) // 2
                half_num_tokens = self.num_tokens // 2
                start = middle - half_num_tokens
                end = start + self.num_tokens
                x = x[:, start:end]
            elif self.location == "last":
                x = x[:, -self.num_tokens:]
            elif self.location == "bilateral":
                num_tokens_half = self.num_tokens // 2
                x = torch.concat([x[:, :num_tokens_half], x[:, -num_tokens_half:]], dim=1)
            elif self.location == "uniform":
                # all but the last chunk are full
                chunk_size = (x.size(1) - self.num_tokens) // (self.num_tokens + 1) + 1
                x = torch.stack([x[:, (i + 1) * chunk_size + i] for i in range(self.num_tokens)], dim=1)
            else:
                raise NotImplementedError

            # aggregate if multiple tokens are used
            if self.aggregate == "flatten":
                return x.flatten(start_dim=1)
            elif self.aggregate == "mean":
                return x.mean(dim=1)
            else:
                raise NotImplementedError



import math

import torch
from torch import nn
from kappamodules.init.functional import (
    init_truncnormal_zero_bias,
    init_xavier_uniform_merged_linear,
    init_zero,
)


class Dit(nn.Module):
    def __init__(
            self,
            cond_dim,
            out_dim,
            num_outputs=6,
            gate_indices=None,
            init_weights="xavier_uniform",
            init_gate_zero=False,
    ):
        super().__init__()
        self.cond_dim = cond_dim
        self.out_dim = out_dim
        self.num_outputs = num_outputs
        self.modulation = nn.Linear(cond_dim, num_outputs * out_dim)
        self.init_weights = init_weights
        self.init_gate_zero = init_gate_zero
        self.gate_indices = gate_indices
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            init_xavier_uniform_merged_linear(self.modulation, num_layers=self.num_outputs)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            init_truncnormal_zero_bias(self.modulation)
        elif self.init_weights == "zero":
            init_zero(self.modulation)
        else:
            raise NotImplementedError
        if self.init_gate_zero:
            assert self.gate_indices is not None
            for gate_index in self.gate_indices:
                start = self.out_dim * gate_index
                end = self.out_dim * (gate_index + 1)
                with torch.no_grad():
                    self.modulation.weight[start:end] = 0
                    self.modulation.bias[start:end] = 0

    def forward(self, cond):
        # exmple: transformer block
        # scale1, shift1, gate1, scale2, shift2, gate2 = self.modulation(cond)
        return self.modulation(cond).chunk(self.num_outputs, dim=1)



import einops
import torch
import torch.nn as nn
import torch.nn.functional as F

from kappamodules.init import init_xavier_uniform_merged_linear
from kappamodules.layers import LinearProjection


class TranssolverAttention(nn.Module):
    """
    Adapted from https://github.com/thuml/Transolver/blob/main/Car-Design-ShapeNetCar/models/Transolver.py
    - readable reshaping operations via einops
    - merged qkv linear layer for higher GPU utilization
    - F.scaled_dot_product_attention instead of slow pytorch attention
    - possibility to mask tokens (required to process variable sized inputs)
    """

    def __init__(
            self,
            dim,
            num_heads,
            num_slices,
            dropout=0.0,
            qkv_bias=False,
            init_weights="truncnormal",
            init_last_proj_zero=False,
    ):
        super().__init__()
        dim_head = dim // num_heads
        self.dim_head = dim_head
        self.num_slices = num_slices
        self.num_heads = num_heads
        self.dropout = dropout
        self.init_last_proj_zero = init_last_proj_zero
        self.temperature = nn.Parameter(torch.full(size=(1, num_heads, 1, 1), fill_value=0.5))

        self.in_project_x = LinearProjection(dim, dim, init_weights=init_weights)
        self.in_project_fx = LinearProjection(dim, dim, init_weights=init_weights)
        self.in_project_slice = LinearProjection(dim_head, num_slices, init_weights=init_weights)
        # this is contained in the original implementation but is useless as it is overwritten later on
        # when initializing all linear layers with truncnormal
        # for l in [self.in_project_slice]:
        #     # use a principled initialization
        #     torch.nn.init.orthogonal_(l.weight)
        self.qkv = LinearProjection(dim_head, dim_head * 3, bias=qkv_bias, init_weights=init_weights)
        self.proj = LinearProjection(dim, dim, init_weights=init_weights)
        self.proj_dropout = nn.Dropout(dropout)

        # init weights
        if init_weights == "xavier_uniform":
            init_xavier_uniform_merged_linear(self.qkv.proj, num_layers=3)
        if self.init_last_proj_zero:
            nn.init.zeros_(self.proj.proj.weight)
            # init_weights == "torch" has no zero bias init
            if self.proj.proj.bias is not None:
                nn.init.zeros_(self.proj.proj.bias)

    def forward(self, x, attn_mask=None):
        batch_size, seqlen, _ = x.shape

        # slice
        fx_mid = einops.rearrange(
            self.in_project_fx(x),
            "batch_size seqlen (num_heads dim_head) -> batch_size num_heads seqlen dim_head",
            num_heads=self.num_heads,
        ).contiguous()
        x_mid = einops.rearrange(
            self.in_project_x(x),
            "batch_size seqlen (num_heads dim_head) -> batch_size num_heads seqlen dim_head",
            num_heads=self.num_heads,
        ).contiguous()
        slice_weights = F.softmax(self.in_project_slice(x_mid) / self.temperature, dim=-1)
        if attn_mask is not None:
            assert attn_mask.dtype == torch.bool, "only bool mask supported"
            assert attn_mask.ndim == 2
            assert len(attn_mask) == len(x)
            assert attn_mask.size(1) == seqlen
            attn_mask = einops.rearrange(attn_mask, "batch_size seqlen -> batch_size 1 seqlen 1").float()
            slice_weights = slice_weights * attn_mask
        slice_norm = einops.rearrange(
            slice_weights.sum(2),
            "batch_size num_heads num_slices -> batch_size num_heads num_slices 1",
        )
        slice_token = torch.einsum("bhnc,bhng->bhgc", fx_mid, slice_weights) / (slice_norm + 1e-5)

        # attention among slice tokens
        q_slice_token, k_slice_token, v_slice_token = self.qkv(slice_token).chunk(3, dim=-1)
        out_slice_token = F.scaled_dot_product_attention(
            q_slice_token,
            k_slice_token,
            v_slice_token,
            dropout_p=self.dropout if self.training else 0.0,
        )

        # deslice
        out_x = torch.einsum("bhgc,bhng->bhnc", out_slice_token, slice_weights)
        out_x = einops.rearrange(
            out_x,
            "batch_size num_heads seqlen dim_head -> batch_size seqlen (num_heads dim_head)",
        )
        out_x = self.proj(out_x)
        out_x = self.proj_dropout(out_x)
        return out_x



import torch.nn.functional as F
from torch import nn

from kappamodules.attention import MMMDiTDotProductAttention
from kappamodules.modulation import Dit
from kappamodules.modulation.functional import modulate_scale_shift, modulate_gate
from kappamodules.utils.param_checking import to_ntuple
from .mlp import Mlp


class MMMDitBlock(nn.Module):
    """
    modular multi-modal adaptive norm block
    adaption of the multi-modal adaptive norm block (https://arxiv.org/abs/2403.03206)
    - supports arbitrary many modalities
    - conditioning is assumed to be per modality
    - interaction is split into "main-branches" and "optional-branches"
        - "main-branches" interact with each other and also propagate gradients through
        - "optional-branches" interact with itself (with gradient) and with the "main-branches" (without gradient)  
    """

    def __init__(
            self,
            dim,
            num_heads,
            num_main_modalities,
            num_optional_modalities,
            cond_dim=None,
            qkv_bias=True,
            act_ctor=nn.GELU,
            eps=1e-6,
            init_weights="truncnormal002",
            init_last_proj_zero=False,
            init_gate_zero=False,
    ):
        super().__init__()
        # preprocess
        num_modalities = num_main_modalities + num_optional_modalities
        dim = to_ntuple(dim, num_modalities)
        # properties
        self.num_modalities = num_modalities
        self.dim = dim
        self.num_heads = num_heads
        self.num_main_modalities = num_main_modalities
        self.num_optional_modalities = num_optional_modalities
        self.eps = eps
        # modulation
        self.modulation = nn.ModuleList(
            [
                Dit(
                    cond_dim=cond_dim,
                    out_dim=dim[i],
                    init_weights=init_weights,
                    num_outputs=6,
                    gate_indices=[2, 5],
                    init_gate_zero=init_gate_zero,
                )
                for i in range(num_modalities)
            ],
        )
        # attn
        assert all(dim[0] == dim[i] for i in range(num_main_modalities)), "main_modalities need same dim"
        self.attn = MMMDiTDotProductAttention(
            dim=dim[0],
            num_heads=num_heads,
            num_main_modalities=num_main_modalities,
            num_optional_modalities=num_optional_modalities,
            qkv_bias=qkv_bias,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        # mlp
        self.mlp = nn.ModuleList(
            [
                Mlp(
                    in_dim=dim[i],
                    hidden_dim=dim[i] * 4,
                    act_ctor=act_ctor,
                    init_weights=init_weights,
                    init_last_proj_zero=init_last_proj_zero,
                )
                for i in range(num_modalities)
            ],
        )

    def _attn_residual_path(self, x, scales, shifts, gates, attn_mask):
        assert isinstance(x, list)
        assert isinstance(scales, list)
        assert isinstance(shifts, list)
        assert isinstance(gates, list)
        assert len(x) == len(scales)
        assert len(x) == len(shifts)
        assert len(x) == len(gates)
        for i in range(len(x)):
            x[i] = F.layer_norm(x[i], [self.dim[i]], eps=self.eps)
            x[i] = modulate_scale_shift(x[i], scale=scales[i], shift=shifts[i])
        x = self.attn(*x, attn_mask=attn_mask)
        for i in range(len(x)):
            x[i] = modulate_gate(x[i], gate=gates[i])
        return x

    def _mlp_residual_path(self, x, scales, shifts, gates):
        assert isinstance(x, list)
        assert isinstance(scales, list)
        assert isinstance(shifts, list)
        assert isinstance(gates, list)
        assert len(x) == len(scales)
        assert len(x) == len(shifts)
        assert len(x) == len(gates)
        for i in range(len(x)):
            x[i] = F.layer_norm(x[i], [self.dim[i]], eps=self.eps)
            x[i] = modulate_scale_shift(x[i], scale=scales[i], shift=shifts[i])
            x[i] = self.mlp[i](x[i])
            x[i] = modulate_gate(x[i], gate=gates[i])
        return x

    def forward(self, x, cond, attn_mask=None):
        assert isinstance(x, (list, tuple))
        assert isinstance(cond, (list, tuple))

        scales_shifts_gates = [self.modulation[i](cond[i]) for i in range(self.num_modalities)]
        scales = [scales_shifts_gates[i][0] for i in range(self.num_modalities)]
        shifts = [scales_shifts_gates[i][1] for i in range(self.num_modalities)]
        gates = [scales_shifts_gates[i][2] for i in range(self.num_modalities)]
        og_x = [x[i] for i in range(len(x))]
        x = self._attn_residual_path(
            x=x,
            scales=scales,
            shifts=shifts,
            gates=gates,
            attn_mask=attn_mask,
        )
        x = [og_x[i] + x[i] for i in range(len(x))]
        scales = [scales_shifts_gates[i][3] for i in range(self.num_modalities)]
        shifts = [scales_shifts_gates[i][4] for i in range(self.num_modalities)]
        gates = [scales_shifts_gates[i][5] for i in range(self.num_modalities)]
        og_x = [x[i] for i in range(len(x))]
        x = self._mlp_residual_path(
            x=x,
            scales=scales,
            shifts=shifts,
            gates=gates,
        )
        x = [og_x[i] + x[i] for i in range(len(x))]
        return x



import einops
import torch
import torch.nn.functional as F
from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)
from kappamodules.utils.param_checking import to_ntuple

class MMMDiTDotProductAttention(nn.Module):
    def __init__(
            self,
            dim,
            num_main_modalities,
            num_optional_modalities,
            num_heads=8,
            qkv_bias=True,
            proj_bias=True,
            channel_first=False,
            init_weights="truncnormal002",
            init_last_proj_zero=False,
    ):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        num_modalities = num_main_modalities + num_optional_modalities
        dim = to_ntuple(dim, num_modalities)
        assert dim[0] % num_heads == 0, "dim[0] should be divisible by num_heads"
        # checks
        self.num_heads = num_heads
        self.num_modalities = num_modalities
        self.num_main_modalities = num_main_modalities
        self.num_optional_modalities = num_optional_modalities
        self.channel_first = channel_first
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero

        # attention is done in main_branch dim
        attn_dim = dim[0]
        self.qkv = nn.ModuleList(
            [
                nn.Linear(dim[i], attn_dim * 3, bias=qkv_bias)
                for i in range(num_modalities)
            ],
        )
        self.proj = nn.ModuleList(
            [
                nn.Linear(attn_dim, dim[i], bias=proj_bias)
                for i in range(num_modalities)
            ],
        )
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            for qkv in self.qkv:
                init_xavier_uniform_merged_linear(qkv, num_layers=3)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            for proj in self.proj:
                nn.init.zeros_(proj.weight)
                # init_weights == "torch" has no zero bias init
                if proj.bias is not None:
                    nn.init.zeros_(proj.bias)

    @staticmethod
    def to_channel_last(x):
        return einops.rearrange(x, "b c l -> b l c")

    @staticmethod
    def to_channel_first(x):
        return einops.rearrange(x, "b l c -> b c l")

    def forward(self, *args, attn_mask=None):
        assert attn_mask is None
        x = list(args)
        assert len(x) == self.num_modalities
        assert all(x[i].ndim == 3 for i in range(self.num_modalities))
        if self.channel_first:
            for i in range(self.num_modalities):
                x[i] = self.to_channel_last(x[i])
        seqlens = [x[i].size(1) for i in range(self.num_modalities)]

        qs = []
        ks = []
        vs = []
        for i in range(self.num_modalities):
            q, k, v = einops.rearrange(
                self.qkv[i](x[i]),
                "bs seqlen (three num_heads head_dim) -> three bs num_heads seqlen head_dim",
                three=3,
                num_heads=self.num_heads,
            ).unbind(0)
            qs.append(q)
            ks.append(k)
            vs.append(v)
        # concat main in sequence dimension
        main_q = torch.concat(qs[:self.num_main_modalities], dim=2)
        main_k = torch.concat(ks[:self.num_main_modalities], dim=2)
        main_v = torch.concat(vs[:self.num_main_modalities], dim=2)
        # main attention
        main_x = F.scaled_dot_product_attention(main_q, main_k, main_v)
        main_x = einops.rearrange(main_x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        main_x = list(main_x.split(seqlens[:self.num_main_modalities], dim=1))

        # optional attention
        optional_x = []
        for i in range(self.num_main_modalities, self.num_modalities):
            q = qs[i]
            k = torch.concat([main_k.detach(), ks[i]], dim=2)
            v = torch.concat([main_v.detach(), vs[i]], dim=2)
            opt_x = F.scaled_dot_product_attention(q, k, v)
            opt_x = einops.rearrange(opt_x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
            optional_x.append(opt_x)

        # postprocess
        x = main_x + optional_x
        for i in range(self.num_modalities):
            x[i] = self.proj[i](x[i])
        if self.channel_first:
            for i in range(self.num_modalities):
                x[i] = self.to_channel_first(x[i])
        return x



import einops
import torch.nn.functional as F
from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)


class LinformerAttention(nn.Module):
    """ DotProductAttention but with an additional projection layer before k and v that reduces sequence length """

    def __init__(
            self,
            dim,
            input_seqlen,
            kv_seqlen,
            num_heads=8,
            qkv_bias=True,
            channel_first=False,
            init_weights="truncnormal002",
            init_last_proj_zero=False,
    ):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.dim = dim
        self.input_seqlen = input_seqlen
        self.kv_seqlen = kv_seqlen
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.channel_first = channel_first
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.to_k_seqlen = nn.Linear(input_seqlen, kv_seqlen)
        self.to_v_seqlen = nn.Linear(input_seqlen, kv_seqlen)
        self.proj = nn.Linear(dim, dim)

        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            init_xavier_uniform_merged_linear(self.qkv, num_layers=3)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            nn.init.zeros_(self.proj.weight)
            # init_weights == "torch" has no zero bias init
            if self.proj.bias is not None:
                nn.init.zeros_(self.proj.bias)

    def to_channel_last(self, x):
        raise NotImplementedError

    def to_channel_first(self, x, og_shape):
        raise NotImplementedError

    def forward(self, x, attn_mask=None):
        assert attn_mask is None
        if self.channel_first:
            og_shape = x.shape
            x = self.to_channel_last(x)
        else:
            og_shape = None

        # to qkv
        q, k, v = einops.rearrange(
            self.qkv(x),
            "bs seqlen (three num_heads head_dim) -> three bs (num_heads head_dim) seqlen",
            three=3,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)
        # reduce kv seqlen
        k = self.to_k_seqlen(k)
        v = self.to_v_seqlen(v)
        # to_k_seqlen and to_v_seqlen expect seqlen at last position -> reshape
        q = einops.rearrange(
            q,
            "bs (num_heads head_dim) seqlen -> bs num_heads seqlen head_dim",
            num_heads=self.num_heads,
        )
        k = einops.rearrange(
            k,
            "bs (num_heads head_dim) seqlen -> bs num_heads seqlen head_dim",
            num_heads=self.num_heads,
        )
        v = einops.rearrange(
            v,
            "bs (num_heads head_dim) seqlen -> bs num_heads seqlen head_dim",
            num_heads=self.num_heads,
        )
        # attention
        x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask)
        x = einops.rearrange(x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        x = self.proj(x)

        if self.channel_first:
            x = self.to_channel_first(x, og_shape=og_shape)
        return x


class LinformerAttention1d(LinformerAttention):
    def to_channel_last(self, x):
        return einops.rearrange(x, "b c l -> b l c")

    def to_channel_first(self, x, og_shape):
        return einops.rearrange(x, "b l c -> b c l")


class LinformerAttention2d(LinformerAttention):
    def to_channel_last(self, x):
        return einops.rearrange(x, "b c h w -> b (h w) c")

    def to_channel_first(self, x, og_shape):
        _, _, h, w = og_shape
        return einops.rearrange(x, "bs (h w) dim -> bs dim h w", h=h, w=w)


class LinformerAttention3d(LinformerAttention):
    def to_channel_last(self, x):
        return einops.rearrange(x, "b c h w d -> b (h w d) c")

    def to_channel_first(self, x, og_shape):
        _, _, h, w, d = og_shape
        return einops.rearrange(x, "bs (h w d) dim -> bs dim h w d", h=h, w=w, d=d)



from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)


class DotProductAttentionSlow(nn.Module):
    def __init__(
            self,
            dim,
            num_heads=8,
            qkv_bias=True,
            qk_norm=False,
            proj_bias=True,
            attn_drop=0.,
            proj_drop=0.,
            norm_layer=nn.LayerNorm,
            init_weights="xavier_uniform",
            init_last_proj_zero=False,
    ):
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim ** -0.5
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)
        self.proj_drop = nn.Dropout(proj_drop)

        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            init_xavier_uniform_merged_linear(self.qkv, num_layers=3)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            nn.init.zeros_(self.proj.weight)
            # init_weights == "torch" has no zero bias init
            if self.proj.bias is not None:
                nn.init.zeros_(self.proj.bias)

    def forward(self, x, attn_mask=None):
        assert attn_mask is None
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q, k = self.q_norm(q), self.k_norm(k)

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = attn @ v

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x



from functools import partial

from torch import nn

from kappamodules.attention import DotProductAttention1d
from kappamodules.init import init_norms_as_noaffine
from kappamodules.layers import DropPath
from kappamodules.modulation import Dit
from kappamodules.modulation.functional import modulate_scale_shift, modulate_gate
from .mlp import Mlp


class DitBlock(nn.Module):
    """ adaptive norm block (https://github.com/facebookresearch/DiT) """

    def __init__(
            self,
            dim,
            num_heads,
            mlp_hidden_dim=None,
            cond_dim=None,
            qkv_bias=True,
            drop_path=0.,
            attn_ctor=DotProductAttention1d,
            eps=1e-6,
            init_weights="xavier_uniform",
            init_norms="nonaffine",
            init_last_proj_zero=False,
            init_modulation_zero=False,
            init_gate_zero=False,
    ):
        super().__init__()
        # DiT uses non-affine LayerNorm and GELU with tanh-approximation
        norm_ctor = partial(nn.LayerNorm, elementwise_affine=False)
        act_ctor = partial(nn.GELU, approximate="tanh")
        # properties
        self.init_norms = init_norms
        mlp_hidden_dim = mlp_hidden_dim or dim * 4
        cond_dim = cond_dim or dim
        # modulation
        self.modulation = Dit(
            cond_dim=cond_dim,
            out_dim=dim,
            init_weights="zero" if init_modulation_zero else init_weights,
            num_outputs=6,
            gate_indices=[2, 5],
            init_gate_zero=init_gate_zero,
        )
        # attn
        self.norm1 = norm_ctor(dim, eps=eps)
        self.attn = attn_ctor(
            dim=dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.drop_path1 = DropPath(drop_prob=drop_path)
        # mlp
        self.norm2 = norm_ctor(dim, eps=eps)
        self.mlp = Mlp(
            in_dim=dim,
            hidden_dim=mlp_hidden_dim,
            act_ctor=act_ctor,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.drop_path2 = DropPath(drop_prob=drop_path)
        #
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_norms == "torch":
            pass
        elif self.init_norms == "nonaffine":
            init_norms_as_noaffine(self.norm1)
            init_norms_as_noaffine(self.norm2)
        else:
            raise NotImplementedError

    def _attn_residual_path(self, x, scale, shift, gate, attn_mask):
        x = modulate_scale_shift(self.norm1(x), scale=scale, shift=shift)
        x = self.attn(x, attn_mask=attn_mask)
        return modulate_gate(x, gate=gate)

    def _mlp_residual_path(self, x, scale, shift, gate):
        return modulate_gate(self.mlp(modulate_scale_shift(self.norm2(x), scale=scale, shift=shift)), gate=gate)

    def forward(self, x, cond, attn_mask=None):
        scale1, shift1, gate1, scale2, shift2, gate2 = self.modulation(cond)
        x = self.drop_path1(
            x,
            residual_path=self._attn_residual_path,
            residual_path_kwargs=dict(
                scale=scale1,
                shift=shift1,
                gate=gate1,
                attn_mask=attn_mask,
            ),
        )
        x = self.drop_path2(
            x,
            residual_path=self._mlp_residual_path,
            residual_path_kwargs=dict(
                scale=scale2,
                shift=shift2,
                gate=gate2,
            ),
        )
        return x



from functools import partial

from torch import nn

from kappamodules.attention import PerceiverAttention1d
from kappamodules.layers import DropPath
from kappamodules.modulation import Dit
from kappamodules.modulation.functional import modulate_scale_shift, modulate_gate
from .mlp import Mlp


class DitPerceiverBlock(nn.Module):
    def __init__(
            self,
            dim,
            num_heads,
            kv_dim=None,
            mlp_hidden_dim=None,
            cond_dim=None,
            drop_path=0.,
            bias=True,
            concat_query_to_kv=False,
            act_ctor=nn.GELU,
            eps=1e-6,
            init_weights="xavier_uniform",
            init_last_proj_zero=False,
            init_modulation_zero=False,
            init_gate_zero=False,
    ):
        super().__init__()
        norm_ctor = partial(nn.LayerNorm, elementwise_affine=False)
        mlp_hidden_dim = mlp_hidden_dim or dim * 4
        cond_dim = cond_dim or dim
        # modulation
        self.modulation = Dit(
            cond_dim=cond_dim,
            out_dim=dim,
            num_outputs=8,
            gate_indices=[4, 7],
            init_weights="zero" if init_modulation_zero else init_weights,
            init_gate_zero=init_gate_zero,
        )
        #
        self.norm1q = norm_ctor(dim, eps=eps)
        self.norm1kv = norm_ctor(kv_dim or dim, eps=eps)
        self.attn = PerceiverAttention1d(
            dim=dim,
            num_heads=num_heads,
            kv_dim=kv_dim,
            bias=bias,
            concat_query_to_kv=concat_query_to_kv,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.drop_path1 = DropPath(drop_prob=drop_path)
        self.norm2 = norm_ctor(dim, eps=eps)
        self.mlp = Mlp(
            in_dim=dim,
            hidden_dim=mlp_hidden_dim,
            bias=bias,
            act_ctor=act_ctor,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.drop_path2 = DropPath(drop_prob=drop_path)

    def _attn_residual_path(self, q, kv, q_scale, q_shift, kv_scale, kv_shift, gate, attn_mask):
        q = modulate_scale_shift(self.norm1q(q), scale=q_scale, shift=q_shift)
        kv = modulate_scale_shift(self.norm1kv(kv), scale=kv_scale, shift=kv_shift)
        x = self.attn(q=q, kv=kv, attn_mask=attn_mask)
        return modulate_gate(x, gate=gate)

    def _mlp_residual_path(self, x, scale, shift, gate):
        return modulate_gate(self.mlp(modulate_scale_shift(self.norm2(x), scale=scale, shift=shift)), gate=gate)

    def forward(self, q, kv, cond, attn_mask=None):
        q_scale, q_shift, kv_scale, kv_shift, attn_gate, mlp_scale, mlp_shift, mlp_gate = self.modulation(cond)
        q = self.drop_path1(
            q,
            residual_path=self._attn_residual_path,
            residual_path_kwargs=dict(
                kv=kv,
                q_scale=q_scale,
                q_shift=q_shift,
                kv_scale=kv_scale,
                kv_shift=kv_shift,
                gate=attn_gate,
                attn_mask=attn_mask,
            ),
        )
        q = self.drop_path2(
            q,
            residual_path=self._mlp_residual_path,
            residual_path_kwargs=dict(
                scale=mlp_scale,
                shift=mlp_shift,
                gate=mlp_gate,
            ),
        )
        return q



import einops
import einops
import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)
from kappamodules.utils.param_checking import to_ntuple


class LocalgridAttention2d(nn.Module):
    def __init__(
            self,
            dim,
            kernel_size=3,
            num_heads=8,
            bias=True,
            init_weights="truncnormal002",
            init_last_proj_zero=False,
    ):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        kernel_size = to_ntuple(kernel_size, n=2)
        padding = [ksize // 2 for ksize in kernel_size]
        assert all(ksize % 2 == 1 for ksize in kernel_size)

        self.kernel_size = kernel_size
        self.padding = padding
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero

        self.qkv = nn.Linear(dim, dim * 3, bias=bias)
        self.relative_position_bias = nn.Parameter(torch.zeros(1, num_heads, 1, np.prod(kernel_size)))
        self.proj = nn.Linear(dim, dim, bias=bias)

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.trunc_normal_(self.relative_position_bias)
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            init_xavier_uniform_merged_linear(self.qkv, num_layers=3)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            nn.init.zeros_(self.proj.weight)
            # init_weights == "torch" has no zero bias init
            if self.proj.bias is not None:
                nn.init.zeros_(self.proj.bias)

    def forward(self, x, attn_mask=None):
        # shapes
        batch_size, seqlen_h, seqlen_w, _ = x.shape
        padding_h, padding_w = self.padding
        kernel_h, kernel_w = self.kernel_size

        # project to attention space
        qkv = self.qkv(x)

        # pad + attention mask
        assert attn_mask is None
        qkv = F.pad(qkv, pad=(0, 0, padding_w, padding_w, padding_h, padding_h), mode="constant", value=0)
        attn_mask = torch.zeros(size=(seqlen_h, seqlen_w), device=x.device)
        attn_mask = F.pad(
            attn_mask,
            pad=(padding_h, padding_h, padding_w, padding_w),
            mode="constant",
            value=float("-inf"),
        )

        # create index
        assert self.kernel_size == (3, 3)
        q_idx = torch.stack(
            torch.meshgrid(
                [
                    torch.arange(padding_h, seqlen_h + padding_h),
                    torch.arange(padding_w, seqlen_w + padding_w),
                ],
                indexing="ij",
            ),
        )
        kv_idx = einops.repeat(q_idx, "ndim seqlen_h seqlen_w -> nine ndim seqlen_h seqlen_w", nine=9).clone()
        # top left
        kv_idx[0] -= 1
        # top center
        kv_idx[1, 0] -= 1
        # top right
        kv_idx[2, 0] -= 1
        kv_idx[2, 1] += 1
        # center left
        kv_idx[3, 1] -= 1
        # center center
        pass
        # center right
        kv_idx[5, 1] += 1
        # bot left
        kv_idx[6, 0] += 1
        kv_idx[6, 1] -= 1
        # bot center
        kv_idx[7, 0] += 1
        # bot right
        kv_idx[8, 0] += 1
        kv_idx[8, 1] += 1

        # flatten indices
        q_idx = q_idx[1] + q_idx[0] * (seqlen_w + padding_h + padding_w)
        q_idx = einops.rearrange(q_idx, "seqlen_h seqlen_w -> (seqlen_h seqlen_w)")
        kv_idx = kv_idx[:, 1] + kv_idx[:, 0] * (seqlen_w + padding_h + padding_w)
        kv_idx = einops.rearrange(kv_idx, "nine seqlen_h seqlen_w -> nine (seqlen_h seqlen_w)")

        # split per head
        q, k, v = einops.rearrange(
            qkv,
            "bs seqlen_h seqlen_w (three num_heads head_dim) -> three bs num_heads (seqlen_h seqlen_w) head_dim",
            three=3,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)

        # gather q
        q_idx_expand = q_idx[None, None, :, None].expand(batch_size, self.num_heads, -1, self.head_dim)
        q = torch.gather(q, dim=2, index=q_idx_expand)
        q = einops.rearrange(q, "batch_size num_heads seqlen head_dim -> (batch_size seqlen) num_heads 1 head_dim")
        # gather kv
        kv_idx = einops.rearrange(kv_idx, "nine seqlen -> (seqlen nine)")
        kv_idx_expand = kv_idx[None, None, :, None].expand(batch_size, self.num_heads, -1, self.head_dim)
        k = torch.gather(k, dim=2, index=kv_idx_expand)
        v = torch.gather(v, dim=2, index=kv_idx_expand)
        k = einops.rearrange(
            k,
            "batch_size num_heads (seqlen nine) head_dim -> (batch_size seqlen) num_heads nine head_dim",
            nine=9,
        )
        v = einops.rearrange(
            v,
            "batch_size num_heads (seqlen nine) head_dim -> (batch_size seqlen) num_heads nine head_dim",
            nine=9,
        )
        # gather mask
        attn_mask = attn_mask.flatten()[kv_idx]
        attn_mask = einops.repeat(
            attn_mask,
            "(seqlen nine) -> (batch_size seqlen) 1 1 nine",
            batch_size=batch_size,
            nine=9,
        )
        # add relative positional bias
        attn_mask = attn_mask + self.relative_position_bias

        x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask)
        x = einops.rearrange(
            x,
            "(bs seqlen_h seqlen_w) num_heads 1 head_dim -> bs seqlen_h seqlen_w (num_heads head_dim)",
            bs=batch_size,
            seqlen_h=seqlen_h,
            seqlen_w=seqlen_w,
        )
        x = self.proj(x)
        return x
